//
//  NSString-ZEditExtension.m
//  ZEdit
//
//  Created by 増田 勇樹 on 05/04/13.
//  Copyright 2005 __MyCompanyName__. All rights reserved.
//

#import "NSString-ZEditExtension.h"
#import <ZScannerKit/ZScannerKit.h>

static NSString *_sharedLineBreakStyleNames[] =
{
	@"Unix (LF)",
	@"Mac (CR)",
	@"Windows (CRLF)",
	@"Unicode Line Separator",
	@"Unicode Paragraph Separator"
};

@interface NSString (ZEditExtensionPrivates)
+ (NSString *)lineBreakStringOfStyle:(NSStringLineBreakStyle)style;
+ (ZScanner *)lineBreakScanner;
@end

@implementation NSString (ZEditExtensionPrivates)
+ (NSString *)lineBreakStringOfStyle:(NSStringLineBreakStyle)style
{
	static unichar linebreaks[4] = {0x000D, 0x000A, 0x2028, 0x2029};
	
	switch(style)
	{
		case NSStringLineBreakCRLF:
			return [NSString stringWithCharacters:&linebreaks[0] length:2];
		case NSStringLineBreakCR:
			return [NSString stringWithCharacters:&linebreaks[0] length:1];
		case NSStringLineBreakLF:
			return [NSString stringWithCharacters:&linebreaks[1] length:1];
		case NSStringLineBreakUnicodeLineSeparator:
			return [NSString stringWithCharacters:&linebreaks[2] length:1];
		case NSStringLineBreakUnicodeParagraphSeparator:
			return [NSString stringWithCharacters:&linebreaks[3] length:1];
		default:
			return nil;
	}
}

+ (ZScanner *)lineBreakScanner
{
	ZScanner *scanner = [ZScanner scanner];
	
	[scanner addRuleWithIdentifier:[NSNumber numberWithInt:NSStringLineBreakCRLF]
			 regularExpression:[NSString lineBreakStringOfStyle:NSStringLineBreakCRLF]];
	[scanner addRuleWithIdentifier:[NSNumber numberWithInt:NSStringLineBreakCR]
			 regularExpression:[NSString lineBreakStringOfStyle:NSStringLineBreakCR]];
	[scanner addRuleWithIdentifier:[NSNumber numberWithInt:NSStringLineBreakLF]
			 regularExpression:[NSString lineBreakStringOfStyle:NSStringLineBreakLF]];
	[scanner addRuleWithIdentifier:[NSNumber numberWithInt:NSStringLineBreakUnicodeLineSeparator]
			 regularExpression:[NSString lineBreakStringOfStyle:NSStringLineBreakUnicodeLineSeparator]];
	[scanner addRuleWithIdentifier:[NSNumber numberWithInt:NSStringLineBreakUnicodeParagraphSeparator]
			 regularExpression:[NSString lineBreakStringOfStyle:NSStringLineBreakUnicodeParagraphSeparator]];
	
	return scanner;
}
@end


@implementation NSString (ZEditExtension)
+ (NSStringEncoding)stringEncodingForLocalizedName:(NSString *)name
{
	const NSStringEncoding *encodings = [NSString availableStringEncodings];
	NSStringEncoding enc;
	while((enc = *encodings++) != 0)
	{
		NSString *n = [NSString localizedNameOfStringEncoding:enc];
		if([n isEqualToString:name])
		{
			return enc;
		}
	}
	return -1;
}

+ (NSArray *)lineBreakStyleNames
{
	return [NSArray arrayWithObjects:_sharedLineBreakStyleNames count:5];
}

+ (NSString *)nameOfLineBreakStyle:(NSStringLineBreakStyle)style
{
	return _sharedLineBreakStyleNames[style];
}

+ (NSStringLineBreakStyle)lineBreakStyleOfName:(NSString *)name
{
	if([name isEqualToString:@"Windows (CRLF)"])
	{
		return NSStringLineBreakCRLF;
	}
	else if([name isEqualToString:@"Mac (CR)"])
	{
		return NSStringLineBreakCR;
	}
	else if([name isEqualToString:@"Unix (LF)"])
	{
		return NSStringLineBreakLF;
	}
	else if([name isEqualToString:@"Unicode Line Separator"])
	{
		return NSStringLineBreakUnicodeLineSeparator;
	}
	else if([name isEqualToString:@"Unicode Paragraph Separator"])
	{
		return NSStringLineBreakUnicodeParagraphSeparator;
	}
	else
	{
		return -1;
	}
}

/* return style of first occurrance of linebrek */
- (NSStringLineBreakStyle)lineBreakStyle
{
	ZScanner *scanner = [NSString lineBreakScanner];
	ZScannerMatch *firstMatch = [[scanner scanEnumeratorInString:self] nextObject];
	if(firstMatch != nil)
	{
		return [[firstMatch identifier] intValue];
	}
	else
	{
		return NSStringLineBreakLF;
	}
}

- (unsigned int)numberOfLines
{
	unsigned int lines = 1;
	ZScanner *scanner = [NSString lineBreakScanner];
	NSEnumerator *matchEnum = [scanner scanEnumeratorInString:self];
	ZScannerMatch *match;
	while((match = [matchEnum nextObject]) != nil)
	{
		lines ++;
	}
	return lines;
}

- (unsigned int)lineNumberAtIndex:(unsigned int)index
{
	unsigned int line = 1;
	ZScanner *scanner = [NSString lineBreakScanner];
	NSEnumerator *matchEnum = [scanner scanEnumeratorInString:self];
	ZScannerMatch *match;
	while((match = [matchEnum nextObject]) != nil)
	{
		if(index < NSMaxRange([match range]))
		{
			break;
		}
		line++;
	}
	return line;
}

- (unsigned int)getLineNumber:(unsigned int *)line atIndex:(unsigned int)index
{
	unsigned int total = 1;
	*line = 1;
	ZScanner *scanner = [NSString lineBreakScanner];
	NSEnumerator *matchEnum = [scanner scanEnumeratorInString:self];
	ZScannerMatch *match;
	while((match = [matchEnum nextObject]) != nil)
	{
		if(index >= NSMaxRange([match range]))
		{
			(*line)++;
		}
		total++;
	}
	return total;
}
@end

@implementation NSMutableString (ZEditExtension)
- (void)setLineBreakStyle:(NSStringLineBreakStyle)style
{
	ZScanner *scanner = [NSString lineBreakScanner];
	NSEnumerator *matchEnum = [scanner scanEnumeratorInString:self];
	ZScannerMatch *match;
	NSString *newLineBreak = [NSString lineBreakStringOfStyle:style];
	int lineBreakLength = [newLineBreak length];
	int indexOffset = 0;
	
	while((match = [matchEnum nextObject]) != nil)
	{
		NSRange realRange = [match range];
		realRange.location += indexOffset;
		[self replaceCharactersInRange:realRange withString:newLineBreak];
		indexOffset += lineBreakLength - [match range].length;
	}
}
@end

