//
//  ZEditTextDocument.m
//  ZEdit
//
//  Created by Youki Masuda on 8.4.2004
//  Copyright (c) 2004 Youki Masuda. All rights reserved.
//

#import "ZEditTextDocument.h"

#import "ZEncodingMenuController.h"
#import "ZMode.h"
#import "ZModeTextView.h"
#import "ZFileSystemItem.h"
#import "ZDictionaryValueTransformer.h"
#import "NSData-ZEditExtention.h"
#import "ZPreferenceController.h"

#define VALUES_APPEND(x) [@"values." stringByAppendingString:x]

@implementation ZEditTextDocument
- (id)init
{
	self = [super init];
	if(self != nil)
	{
		NSUserDefaultsController *udc = [NSUserDefaultsController sharedUserDefaultsController];
		_encoding = [[[udc values] valueForKeyPath:DEFAULT_ENCODING_KEY] intValue];
	}
	return self;
}

- (void)dealloc
{
	if(_textView != nil)
	{
		[_textView setDelegate:nil];
		[_textView unbind:@"font"];
		[_textView unbind:@"colors"];
		[_textView unbind:@"shouldAntialias"];
	}
	[_textView release];
	[super dealloc];
}

- (void)setLineBreakStyle:(NSStringLineBreakStyle)style
{
	_lineBreakStyle = style;
}

- (NSStringLineBreakStyle)lineBreakStyle
{
	return _lineBreakStyle;
}

- (IBAction)changeLineBreakStyle:(id)sender
{
	[self setLineBreakStyle:[NSString lineBreakStyleOfName:[sender title]]];
}


- (void)setEncoding:(NSStringEncoding)encoding
{
	if(_encoding != encoding)
	{
		if([[_textView textStorage] length] > 0 && [self fileURL] != nil)
		{
			NSNumber *encodingNumber = [[NSNumber numberWithInt:encoding] retain];
			NSBeginAlertSheet(NSLocalizedString(@"ZEditChangeEncodingTitle", nil),
							  NSLocalizedString(@"ZEditConvertEncodingButton", nil),
							  NSLocalizedString(@"ZEditReinterpretButton", nil),
							  NSLocalizedString(@"ZEditCancelButton", nill),
							  [self window],
							  self,
							  @selector(sheetDidEnd:returnCode:contextInfo:),
							  nil,
							  encodingNumber,
							  NSLocalizedString(@"ZEditChangeEncodingMessage", nil));
		}
		else
		{
			[self convertToEncoding:encoding];
		}
	}
}

- (void)convertToEncoding:(NSStringEncoding)encoding
{
	if(_encoding != encoding)
	{
		[[self undoManager] setActionName:@"Convert Encoding"];
		[[[self undoManager] prepareWithInvocationTarget:self] convertToEncoding:_encoding];
		_encoding = encoding;
	}
}

- (void)sheetDidEnd:(NSWindow *)sheet returnCode:(int)returnCode contextInfo:(void *)contextInfo
{
	NSStringEncoding encoding = [(NSNumber *)contextInfo intValue];
	[(NSNumber *)contextInfo release];
	if(returnCode == NSAlertDefaultReturn)
	{
		[self convertToEncoding:encoding];
	}
	else if(returnCode == NSAlertAlternateReturn)
	{
		NSError *error = nil;
		NSArray *encodings = [NSArray arrayWithObject:[NSNumber numberWithInt:encoding]];
		
		NSData *data = [NSData dataWithContentsOfURL:[self fileURL] options:0 error:&error];
		if(data != nil && [self readFromData:data withEncodings:encodings error:&error])
		{
			_encoding = encoding;
			[[self undoManager] removeAllActions];
		}
		else
		{
			[self presentError:error];
		}
	}
}

- (NSStringEncoding)encoding
{
	return _encoding;
}

- (IBAction)changeEncoding:(id)sender
{
	[self setEncoding:[sender tag]];
}

- (ZModeTextView *)view
{
	if(_textView == nil)
	{
		_textView = [[ZModeTextView alloc] initWithFrame:NSZeroRect];
		[_textView setRichText:NO];
		[_textView setContinuousSpellCheckingEnabled:NO];
		[_textView setHorizontallyResizable:YES];
		[_textView setUsesFontPanel:NO];
		[_textView setUsesRuler:NO];
		[_textView setUsesFindPanel:YES];
		[_textView setAutoresizingMask:NSViewWidthSizable|NSViewHeightSizable];
		
		[_textView setMode:[ZMode modeForFile:[self displayName]]];
		NSValueTransformer *unarchiver = [NSValueTransformer valueTransformerForName:@"NSUnarchiveFromData"];
		ZDictionaryValueTransformer *colorDictTransformer = [ZDictionaryValueTransformer transFormerWithTransformer:unarchiver];
		[_textView bind:@"font"
				   toObject:[NSUserDefaultsController sharedUserDefaultsController]
				   withKeyPath:VALUES_APPEND(PREFERENCE_FONT_KEY)
				   options:[NSDictionary dictionaryWithObject:@"ZFontUnarchiver" forKey:NSValueTransformerNameBindingOption]];
		[_textView bind:@"shouldAntialias"
				   toObject:[NSUserDefaultsController sharedUserDefaultsController]
				   withKeyPath:VALUES_APPEND(PREFERENCE_ANTI_ALIAS_KEY)
				   options:nil];
		[_textView bind:@"colors"
				   toObject:[NSUserDefaultsController sharedUserDefaultsController]
				   withKeyPath:VALUES_APPEND(PREFERENCE_COLORS_KEY)
				   options:[NSDictionary dictionaryWithObject:colorDictTransformer forKey:NSValueTransformerBindingOption]];
		
		[_textView setDelegate:self];
	}
	return _textView;
}

- (NSTextStorage *)textStorage
{
	return [[self view] textStorage];
}

- (BOOL)readFromData:(NSData *)data ofType:(NSString *)typeName error:(NSError **)outError
{
	NSStringEncoding encoding = [data stringEncoding];
	if(encoding != 0)
	{
		NSString *newText = [[[NSString alloc] initWithData:data encoding:encoding] autorelease];
		if(newText != nil)
		{
			[[self view] setString:newText];
			_encoding = encoding;
			return YES;
		}
	}
	NSUserDefaultsController *udc = [NSUserDefaultsController sharedUserDefaultsController];
	NSNumber *defaultEncoding = [[udc values] valueForKeyPath:DEFAULT_ENCODING_KEY];
	NSMutableArray *encodings = [[[udc values] valueForKeyPath:AVAILABLE_ENCODINGS_KEY] mutableCopy];
	[encodings removeObject:defaultEncoding];
	[encodings insertObject:defaultEncoding atIndex:0];
	return [self readFromData:data withEncodings:encodings error:outError];
}

- (BOOL)readFromData:(NSData *)data withEncodings:(NSArray *)encodings error:(NSError **)outError
{
	NSString *newText = nil;
	int i;
	for(i = 0; i < [encodings count]; i++)
	{
		NSStringEncoding enc = [[encodings objectAtIndex:i] intValue];
		newText = [[[NSString alloc] initWithData:data encoding:enc] autorelease];
		if(newText != nil)
		{
			_encoding = enc;
			break;
		}
	}
	
	if(newText != nil)
	{
		[[self view] setString:newText];
		return YES;
	}
	else
	{
		*outError = [NSError errorWithDomain:NSCocoaErrorDomain code:NSFileReadInapplicableStringEncodingError userInfo:nil];
		return NO;
	}
}

- (NSData *)dataOfType:(NSString *)typeName error:(NSError **)outError
{
	[_textView breakUndoCoalescing];
	NSMutableString *string = [NSMutableString stringWithString:[[self textStorage] string]];
	[string setLineBreakStyle:[self lineBreakStyle]];
	NSData *data = [string dataUsingEncoding:_encoding allowLossyConversion:NO];
	if(data != nil)
	{
		return data;
	}
	else
	{
		*outError = [NSError errorWithDomain:NSCocoaErrorDomain code:NSFileWriteInapplicableStringEncodingError userInfo:nil];
		return NO;
	}
}

- (BOOL)validateUserInterfaceItem:(id <NSValidatedUserInterfaceItem>)anItem
{
	if([anItem action] == @selector(changeLineBreakStyle:))
	{
		NSStringLineBreakStyle currentStyle = [self lineBreakStyle];
		NSStringLineBreakStyle menuStyle = [anItem tag];
		
		if(currentStyle == menuStyle)
		{
			if([(NSObject *)anItem isKindOfClass:[NSMenuItem class]])
			{
				[(NSMenuItem *)anItem setState:NSOnState];
			}
			return NO;
		}
		else
		{
			if([(NSObject *)anItem isKindOfClass:[NSMenuItem class]])
			{
				[(NSMenuItem *)anItem setState:NSOffState];
			}
			return YES;
		}
		return YES;
	}
	else if([anItem action] == @selector(changeEncoding:))
	{
		NSStringEncoding currentEncoding = [self encoding];
		NSStringEncoding menuEncoding = [anItem tag];
		
		if(currentEncoding == menuEncoding)
		{
			if([(NSObject *)anItem isKindOfClass:[NSMenuItem class]])
			{
				[(NSMenuItem *)anItem setState:NSOnState];
			}
			return NO;
		}
		else
		{
			if([(NSObject *)anItem isKindOfClass:[NSMenuItem class]])
			{
				[(NSMenuItem *)anItem setState:NSOffState];
			}
			return YES;
		}
		return YES;
	}
	else
	{
		return [super validateUserInterfaceItem:anItem];
	}
}

// printing
- (NSPrintOperation *)printOperationWithSettings:(NSDictionary *)printSettings error:(NSError **)outError
{
	NSPrintInfo *info = [[[NSPrintInfo alloc] initWithDictionary:printSettings] autorelease];
	[info setVerticallyCentered:NO];
	[info setHorizontalPagination:NSFitPagination];
	NSPrintOperation *op = [NSPrintOperation printOperationWithView:[self view]
											 printInfo:info];
	return op;
}
@end
