#import "ZFileSystemItem.h"

#import "NSArray-ZEditExtension.h"
#import "DRFileNotificationSubscription.h"

static NSArray *_invisibleFiles = nil;

NSArray *InvisibleFiles();
NSDictionary *FileAttributes(NSString *path);
BOOL FileIsDirectory(NSString *path);
BOOL FileIsSymbolicLink(NSString *path);
BOOL FileIsAlias(NSString *path);
BOOL FileIsPackage(NSString *path);
BOOL FileIsInvisible(NSString *path);
NSString *findUniqueName(NSString *name, NSArray *usedNames);

@interface ZFileSystemItem (Privates)
- (NSString *)parentPath;
- (NSString *)fileNameFromDisplayName:(NSString *)displayName;
- (void)insertObject:(ZFileSystemItem *)item inChildrenAtIndex:(unsigned int)index;
- (void)removeObjectFromChildrenAtIndex:(unsigned int)index;
- (void)directoryContentsChanged:(id)sender;
- (BOOL)performOperation:(NSString *)operation destination:(NSString *)distination;
- (void)updateChildren;
@end

@implementation ZFileSystemItem
+ (id)rootItemWithPath:(NSString *)path
{
	return [[[self alloc] initAsRootWithPath:path] autorelease];
}

- (id)initAsRootWithPath:(NSString *)path
{
	return [self initWithPath:path parent:nil owner:nil];
}

- (id)initAsRootWithPath:(NSString *)path owner:(ZFileSystemItem *)owner
{
	return [self initWithPath:path parent:nil owner:owner];
}

- (id)initWithPath:(NSString *)path parent:(ZFileSystemItem *)parent
{
	return [self initWithPath:path parent:parent owner:parent];
}

- (id)initWithPath:(NSString *)path parent:(ZFileSystemItem *)parent owner:(ZFileSystemItem *)owner
{
	self = [super init];
	if(self != nil)
	{
		path = [path stringByStandardizingPath];
		if(![[NSFileManager defaultManager] fileExistsAtPath:path])
		{
			[self release];
			self = nil;
			return self;
		}
        if(parent == nil)
		{
			_parentPath = [[path stringByDeletingLastPathComponent] retain];
		}
		else
		{
			_parent = parent;
		}
		NSString *lastComp = [path lastPathComponent];
		if([lastComp isEqualToString:@"/"])
		{
			_fileName = [@"" retain];
		}
		else
		{
			_fileName = [lastComp retain];
		}
		
		_owner = owner;
		if(FileIsDirectory(path))
		{
			_flags |= FileSystemItemIsDirectory;
			if(FileIsPackage(path))
			{
				_flags |= FileSystemItemIsPackage;
			}
		}
		else 
		{
			if(FileIsSymbolicLink(path))
			{
				_flags |= FileSystemItemIsSymbolicLink;
			}
			else if(FileIsAlias(path))
			{
				_flags |= FileSystemItemIsAlias;
			}
			if(FileIsInvisible(path))
			{
				_flags |= FileSystemItemIsInvisible;
			}
		}
		if([self isDirectory])
		{
			_directoryNotification = [[DRFileNotificationSubscription alloc] initForDirectory:[self path] target:self action:@selector(directoryContentsChanged:)];
		}
    }
    return self;
}

- (id)copyWithZone:(NSZone *)zone
{
	return [[[self class] rootItemWithPath:[self path]] retain];
}

- (void)dealloc
{
	[_directoryNotification release];
	[_resolvedItem release];
    [_children release];
	[_icon release];
    [_parentPath release];
	[_fileName release];
    [super dealloc];
}

- (BOOL)isEqual:(ZFileSystemItem *)item
{
	return [[self path] isEqualToString:[item path]] &&
		   [self typeFlags] == [item typeFlags];
}

- (unsigned)hash
{
	return [[self path] hash] + [self typeFlags];
}

- (NSString *)path
{
	NSString *parentPath = [self parentPath];
	NSString *fileName = [self fileName];
	
	return [parentPath stringByAppendingPathComponent:fileName];
}

- (BOOL)readable
{
	return [[NSFileManager defaultManager] isReadableFileAtPath:[self path]];
}

- (BOOL)writable
{
	return [[NSFileManager defaultManager] isWritableFileAtPath:[self path]];
}

- (BOOL)executable
{
	return [[NSFileManager defaultManager] isExecutableFileAtPath:[self path]];
}

- (NSImage *)icon
{
	if(_icon == nil)
	{
		_icon = [[[NSWorkspace sharedWorkspace] iconForFile:[self path]] retain];
	}
	return _icon;
}

- (NSDictionary *)fileAttributes
{
	return [[NSFileManager defaultManager] fileAttributesAtPath:[self path] traverseLink:NO];
}

- (void)setExtensionHidden:(BOOL)hidden
{
	if([self extensionHidden] != hidden)
	{
		NSDictionary *att = [NSDictionary dictionaryWithObject:[NSNumber numberWithBool:hidden] forKey:NSFileExtensionHidden];
		[[NSFileManager defaultManager] changeFileAttributes:att atPath:[self path]];
	}
}

- (BOOL)extensionHidden
{
	return [[[self fileAttributes] objectForKey:NSFileExtensionHidden] boolValue];
}

- (FileSystemItemTypeFlags)typeFlags
{
	return _flags;
}

- (BOOL)isDirectory
{
	return _flags & FileSystemItemIsDirectory;
}

- (BOOL)isSymbolicLink
{
	return _flags & FileSystemItemIsSymbolicLink;
}

- (BOOL)isAlias
{
	return _flags & FileSystemItemIsAlias;
}

- (BOOL)isPackage
{
	return _flags & FileSystemItemIsPackage;
}

- (BOOL)isInvisible
{
	return _flags & FileSystemItemIsInvisible;
}

- (ZFileSystemItem *)resolveIfSymbolicLinkOrAlias
{
	if([self isSymbolicLink])
	{
		return [self resolveSymbolicLink];
	}
	else if([self isAlias])
	{
		return [self resolveAlias];
	}
	else
	{
		return self;
	}
}

- (ZFileSystemItem *)resolveSymbolicLink
{
	if(_resolvedItem == nil && [self isSymbolicLink])
	{
		NSFileManager *fileManager = [NSFileManager defaultManager];
		NSString *resolvedPath = [fileManager pathContentOfSymbolicLinkAtPath:[self path]];
		_resolvedItem = [[[self class] alloc] initAsRootWithPath:resolvedPath owner:self];
	}
	return _resolvedItem;
}

- (ZFileSystemItem *)resolveAlias
{
	if(_resolvedItem == nil)
	{
		if([self isAlias])
		{
			FSRef fsRef;
			CFURLRef urlRef;
			urlRef = CFURLCreateWithFileSystemPath(NULL, (CFStringRef)[self path], kCFURLPOSIXPathStyle, NO);
			if(CFURLGetFSRef(urlRef, &fsRef))
			{
				Boolean isFolder, wasAliased;
				if(FSResolveAliasFile(&fsRef, true, &isFolder, &wasAliased) == noErr && wasAliased)
				{
					CFURLRef resolvedUrl = CFURLCreateFromFSRef(NULL, &fsRef);
					if(resolvedUrl != nil)
					{
						NSString *resolvedPath = (NSString *)CFURLCopyFileSystemPath(resolvedUrl, kCFURLPOSIXPathStyle);
						CFRelease(resolvedUrl);
						_resolvedItem = [[[self class] alloc] initAsRootWithPath:resolvedPath owner:self];
					}
				}
			}
			CFRelease(urlRef);
		}
	}
	return _resolvedItem;
}

- (void)setIncludesInvisibleChildren:(BOOL)include
{
	_includesInvisibleChildren = include;
}

- (BOOL)includesInvisibleChildren
{
	if(_includesInvisibleChildren)
	{
		return YES;
	}
	else if(_owner != nil)
	{
		return [_owner includesInvisibleChildren];
	}
	else
	{
		return NO;
	}
}

- (void)setTreatsFilePackagesAsDirectories:(BOOL)asDirectory
{
	_treatFilePackagesAsDirectories = asDirectory;
}

- (BOOL)treatsFilePackagesAsDirectories
{
	if(_treatFilePackagesAsDirectories)
	{
		return YES;
	}
	else if(_owner != nil)
	{
		return [_owner treatsFilePackagesAsDirectories];
	}
	else
	{
		return NO;
	}
}

- (BOOL)canRename
{
	NSString *path = [self path];
	if([path isEqualToString:@"/"])
	{
		return NO;
	}
	if([[[NSWorkspace sharedWorkspace] mountedLocalVolumePaths] containsObject:path])
	{
		return NO;
	}
	if([[[NSWorkspace sharedWorkspace] mountedRemovableMedia] containsObject:path])
	{
		return NO;
	}
	return [self writable] && _parent != nil && [_parent writable];
}

- (BOOL)canRenameTo:(NSString *)fileName
{
	if(![self canRename])
	{
		return NO;
	}
	if([fileName isEqualToString:[self fileName]])
	{
		return YES;
	}
	else if([fileName isEqualToString:@""])
	{
		return NO;
	}
	else
	{
		NSCharacterSet *invalidCharacters = [NSCharacterSet characterSetWithCharactersInString:@":/"];
		NSRange range = [fileName rangeOfCharacterFromSet:invalidCharacters];
		return (range.location == NSNotFound);
	}
}

- (void)setFileName:(NSString *)fileName
{
	if(![fileName isEqualToString:[self fileName]])
	{
		NSString *path = [self path];
		NSString *newPath = [[self parentPath] stringByAppendingPathComponent:fileName];
		
		if([[NSFileManager defaultManager] movePath:path toPath:newPath handler:nil])
		{
			[_fileName release];
			_fileName = [fileName copy];
		}
	}
}

- (NSString *)fileName
{
	return _fileName;
}

- (BOOL)canChangeDisplayNameTo:(NSString *)newDisplayName
{
	return [self canRename] && [self canRenameTo:[self fileNameFromDisplayName:newDisplayName]];
}

- (void)setDisplayName:(NSString *)newDisplayName
{
	BOOL dispHasExt = ![[newDisplayName pathExtension] isEqualToString:@""];
	NSString *newFileName = [self fileNameFromDisplayName:newDisplayName];
	BOOL fileHasExt = ![[newFileName pathExtension] isEqualToString:@""];
	
	if(dispHasExt)
	{
		[self setExtensionHidden:NO];
	}
	else
	{
		if(fileHasExt)
		{
			[self setExtensionHidden:YES];
		}
	}
	
	[self setFileName:newFileName];
}

- (NSString *)displayName
{
	return [[NSFileManager defaultManager] displayNameAtPath:[self path]];
}

- (BOOL)moveToDirectory:(NSString *)directoryPath
{	
	return [self performOperation:NSWorkspaceMoveOperation destination:directoryPath];
}

- (BOOL)moveToTrash
{
	return [self performOperation:NSWorkspaceRecycleOperation destination:@""];
}

- (NSString *)nameForUntitledChildFile
{
	NSString *name = NSLocalizedString(@"ZEditUntitledFileName", nil);
	NSArray *childNames = [[self children] valueForKey:@"fileName"];
	
	return findUniqueName(name, childNames);
}

- (NSString *)nameForUntitledChildDirectory
{
	NSString *name = NSLocalizedString(@"ZEditUntitledDirectoryName", nil);
	NSArray *childNames = [[self children] valueForKey:@"fileName"];
	
	return findUniqueName(name, childNames);
}

- (ZFileSystemItem *)createChildDirectoryNamed:(NSString *)directoryName
{
	ZFileSystemItem *parent = [self resolveIfSymbolicLinkOrAlias];
	if([parent isDirectory])
	{
		NSString *directoryPath = [[parent path] stringByAppendingPathComponent:directoryName];
		if([[NSFileManager defaultManager] createDirectoryAtPath:directoryPath attributes:nil])
		{
			ZFileSystemItem *new = [[[[self class] alloc] initWithPath:directoryPath parent:parent] autorelease];
			[parent insertObject:new inChildrenAtIndex:[[self children] count]];
			if(parent != self)
			{
				[self willChangeValueForKey:@"children"];
				[self didChangeValueForKey:@"children"];
			}
			return new;
		}
	}
	return nil;
}

- (ZFileSystemItem *)createChildFileNamed:(NSString *)fileName
{
	ZFileSystemItem *parent = [self resolveIfSymbolicLinkOrAlias];
	if([parent isDirectory])
	{
		NSString *filePath = [[parent path] stringByAppendingPathComponent:fileName];
		if([[NSFileManager defaultManager] createFileAtPath:filePath contents:[NSData data] attributes:nil])
		{
			ZFileSystemItem *new = [[[[self class] alloc] initWithPath:filePath parent:parent] autorelease];
			[parent insertObject:new inChildrenAtIndex:[[self children] count]];
			if(parent != self)
			{
				[self willChangeValueForKey:@"children"];
				[self didChangeValueForKey:@"children"];
			}
			return new;
		}
	}
	return nil;
}

- (ZFileSystemItem *)parent
{
	return _parent;
}

- (NSArray *)children
{
	if(_children == nil)
	{
		if([self isDirectory] && (![self isPackage] || [self treatsFilePackagesAsDirectories]))
		{
			NSFileManager *fileManager = [NSFileManager defaultManager];
			NSString *path = [self path];
			NSArray *cs = [fileManager directoryContentsAtPath:path];
			int i;
			_children = [[NSMutableArray alloc] initWithCapacity:[cs count]];
			for (i = 0; i < [cs count]; i++)
			{
				NSString *fullPath = [self path];
				fullPath = [fullPath stringByAppendingPathComponent:[cs objectAtIndex:i]];
				if([self includesInvisibleChildren] || !FileIsInvisible(fullPath))
				{
					ZFileSystemItem *newItem = [[[[self class] alloc] initWithPath:fullPath parent:self] autorelease];
					if(newItem != nil)
					{
						[_children addObject:newItem];
					}
				}
			}
		}
		else if([self isSymbolicLink])
		{
			return [[self resolveSymbolicLink] children];
		}
		else if([self isAlias])
		{
			return [[self resolveAlias] children];
		}
		else
		{
			return nil;
		}
	}
	return _children;
}

- (int)countOfChildren
{
	return [[self children] count];
}

- (ZFileSystemItem *)objectInChildrenAtIndex:(unsigned int)index
{
	return [[self children] objectAtIndex:index];
}

- (NSIndexPath *)indexPathFromRoot
{
	if([self parent] != nil)
	{
		int myIndex = [[[self parent] children] indexOfObject:self];
		return [[[self parent] indexPathFromRoot] indexPathByAddingIndex:myIndex];
	}
	else if(_owner != nil)
	{
		return [_owner indexPathFromRoot];
	}
	else
	{
		return [NSIndexPath indexPathWithIndex:0];
	}
}

- (BOOL)isLeaf
{
	ZFileSystemItem *item = [self resolveIfSymbolicLinkOrAlias];
	return ![item isDirectory];
}
@end


@implementation ZFileSystemItem (Privates)
- (NSString *)parentPath
{
	if(_parentPath == nil)
	{
		return [_parent path];
	}
	else
	{
		return _parentPath;
	}
}

- (NSString *)fileNameFromDisplayName:(NSString *)displayName
{
	NSString *displayNameExt = [displayName pathExtension];
	if([displayNameExt isEqualToString:@""])
	{
		if(![displayName isEqualToString:@""] && ![[[self fileName] pathExtension] isEqualToString:@""])
		{
			return [displayName stringByAppendingPathExtension:[[self fileName] pathExtension]];
		}
	}
	return displayName;
}

- (void)insertObject:(ZFileSystemItem *)item inChildrenAtIndex:(unsigned int)index
{
	if([self isDirectory])
	{
		if(_children != nil)
		{
			[_children insertObject:item atIndex:index];
		}
	}
}

- (void)removeObjectFromChildrenAtIndex:(unsigned int)index
{
	if([self isDirectory])
	{
		if(_children != nil)
		{
			[_children removeObjectAtIndex:index];
		}
	}
}

- (void)directoryContentsChanged:(id)sender
{
	[self updateChildren];
	if(_parent == nil && _owner != nil)
	{
		[_owner updateChildren];
	}
}

- (BOOL)performOperation:(NSString *)operation destination:(NSString *)distination
{
	NSString *source = [[self path] stringByDeletingLastPathComponent];
	NSArray *files = [NSArray arrayWithObject:[self fileName]];
	int tag = 0;
	return [[NSWorkspace sharedWorkspace] performFileOperation:operation source:source destination:distination files:files tag:&tag];
}

- (void)updateChildren
{
	if([self isDirectory])
	{
		[self willChangeValueForKey:@"children"];
		if(_children != nil)
		{
			NSMutableArray *savedChildren = [[_children mutableCopy] autorelease];
			
			[_children release];
			_children = nil;
			
			NSMutableArray *new = [[[self children] mutableCopy] autorelease];
			
			NSMutableArray *removed = [[savedChildren mutableCopy] autorelease];
			[removed removeObjectsInArray:new];
			
			[new removeObjectsInArray:savedChildren];
			
			[savedChildren addObjectsFromArray:new];
			[savedChildren removeObjectsInArray:removed];
			
			int i;
			for(i = 0; i < [removed count]; i++)
			{
				ZFileSystemItem *item = [removed objectAtIndex:i];
				[[NSNotificationCenter defaultCenter] postNotificationName:FileSystemItemRemovedFromTreeNotification object:item];
			}
			
			[_children release];
			_children = [savedChildren retain];
		}		
		[self didChangeValueForKey:@"children"];
	}
	else if([self isAlias] || [self isSymbolicLink])
	{
		[self willChangeValueForKey:@"children"];
		[self didChangeValueForKey:@"children"];
	}
}
@end

NSArray *InvisibleFiles()
{
	if(_invisibleFiles == nil)
	{
		_invisibleFiles = [[NSArray arrayWithObjects:@"/mach",
													@"/mach.sym",
													@"/tmp",
													@"/var",
													@"/etc",
													@"/Network",
													nil] retain];
	}
	return _invisibleFiles;
}

NSDictionary *FileAttributes(NSString *path)
{
	return [[NSFileManager defaultManager] fileAttributesAtPath:path traverseLink:NO];	
}

BOOL FileIsDirectory(NSString *path)
{
	BOOL isDirectory;
	[[NSFileManager defaultManager] fileExistsAtPath:path isDirectory:&isDirectory];
	return isDirectory;
}

BOOL FileIsSymbolicLink(NSString *path)
{
	return [[FileAttributes(path) objectForKey:NSFileType] isEqualToString:NSFileTypeSymbolicLink];
}

BOOL FileIsAlias(NSString *path)
{
	Boolean isAlias = NO;
	FSRef fsRef;
	CFURLRef urlRef;
	urlRef = CFURLCreateWithFileSystemPath(NULL, (CFStringRef)path, kCFURLPOSIXPathStyle, NO);
	if(CFURLGetFSRef(urlRef, &fsRef))
	{
		Boolean isDirectory;
		FSIsAliasFile(&fsRef, &isAlias, &isDirectory);
	}
	CFRelease(urlRef);
	return isAlias == true;
}

BOOL FileIsPackage(NSString *path)
{
	CFURLRef url = CFURLCreateWithFileSystemPath(NULL, (CFStringRef)path, kCFURLPOSIXPathStyle, false);

	LSItemInfoRecord info;
	if(LSCopyItemInfoForURL(url, kLSRequestAllInfo, &info) == noErr)
	{
		CFRelease(url);
		return 0 != (info.flags & kLSItemInfoIsPackage);
	}
	return NO;
}

BOOL FileIsInvisible(NSString *path)
{
	if([[path lastPathComponent] hasPrefix:@"."])
	{
		return YES;
	}
	if([InvisibleFiles() containsObject:path])
	{
		return YES;
	}
	BOOL isDirectory;
	[[NSFileManager defaultManager] fileExistsAtPath:path isDirectory:&isDirectory];
	FSRef fsRef;
	CFURLRef urlRef;
	urlRef = CFURLCreateWithFileSystemPath(NULL, (CFStringRef)path, kCFURLPOSIXPathStyle, isDirectory);
	if(CFURLGetFSRef(urlRef, &fsRef))
	{
		FSCatalogInfo catalogInfo;
		HFSUniStr255 name;
		if(FSGetCatalogInfo(&fsRef, kFSCatInfoFinderInfo, &catalogInfo, &name, NULL, NULL) == noErr)
		{
			FInfo *finderInfo = (FInfo *)&(catalogInfo.finderInfo);
			CFRelease(urlRef);
			return ((finderInfo->fdFlags)&kIsInvisible) !=0;
		}
	}
	CFRelease(urlRef);
	return YES;
}

NSString *findUniqueName(NSString *name, NSArray *usedNames)
{
	int i;
	for(i = 1; YES;i++)
	{
		NSString *tmpName;
		if(i == 1)
		{
			tmpName = name;
		}
		else
		{
			NSString *extension = [name pathExtension];
			if(extension != nil && ![extension isEqualToString:@""])
			{
				tmpName = [[[name stringByDeletingPathExtension] stringByAppendingFormat:@" %d", i] stringByAppendingPathExtension:extension];
			}
			else
			{
				tmpName = [NSString stringWithFormat:@"%@ %d", name, i];
			}
		}
		
		if(![usedNames logicalOrUnsingSelector:@selector(isEqualToString:) argument:tmpName])
		{
			name = tmpName;
			break;
		}
	}
	return name;
}

